# -*- coding: utf-8 -*-
"""
Routines for parsing LEMI field spreadsheets(i.e. install and demobilization
sheets) and validating metadata;
Templates for the spreadsheets are provided to the PIs and their formatting is
hard coded.
The metadata for a complete  MT dataset was structured into the following
categories: Survey, Station, Run, Electric, Magnetic, and Auxiliary

Maeva Pourpoint - IRIS/PASSCAL
"""
from __future__ import annotations

import copy
import logging
import logging.config
import openpyxl
import re

from configobj import ConfigObj
from dataclasses import dataclass, field, fields
from datetime import datetime
from obspy import UTCDateTime
from pathlib import Path
from typing import Dict, List, Optional, Set, Union, TYPE_CHECKING

from lemi2seed.lemi_data import (SAMPLING_RATE, VALID_COMPONENTS,
                                 CHANNEL_NAMING_CONVENTION)
from lemi2seed.utils import (check_email_formatting, check_instrument_specs,
                             check_serial_number, get_run_list, is_empty,
                             is_valid_uri, get_e_ids, str2list, ELECTRODE_DEFAULT,
                             FLUXGATE_DEFAULT, MSG_E_CHA, NBR_E_CHANNEL_MAX)

if TYPE_CHECKING:
    import numpy
    from lemi2seed.lemi_data import LemiData

# Read lemi2seed config file
config_file = Path(__file__).parent.resolve().joinpath('config.ini')
config = ConfigObj(str(config_file))

# Read logging config file
log_file_path = Path(__file__).parent.resolve().joinpath('logging.conf')
logging.config.fileConfig(log_file_path)
# Create logger
logger = logging.getLogger(__name__)

E_COMPONENT_NAMING_CONVENTION = {'North-South': 'Ex', 'South-North': 'Ex',
                                 'East-West': 'Ey', 'West-East': 'Ey'}


@dataclass
class BaseMetadata:
    """
    This class handles the metadata attributes shared at all levels: survey;
    station; run; electric; magnetic; auxiliary.
    """
    # List of metadata fields that are required for archiving but "missing"
    # because not provided by the user.
    metadata_missing: Set[str] = field(default_factory=set)
    # List of metadata fields that are invalid
    metadata_invalid: Set[str] = field(default_factory=set)


@dataclass
class BaseSurvey(BaseMetadata):
    """
    This class handles the metadata attributes shared at the survey and station
    levels.
    """
    restricted_status: Optional[str] = field(default=None,
                                             metadata={'xml_id': ['restricted_status'],
                                                       'required': True,
                                                       'gui': True})
    time_period_end: Optional[UTCDateTime] = field(default=None,
                                                   metadata={'xml_id': ['end_date'],
                                                             'required': True,
                                                             'gui': True})
    time_period_start: Optional[UTCDateTime] = field(default=None,
                                                     metadata={'xml_id': ['start_date'],
                                                               'required': True,
                                                               'gui': True})

    def validate_time_period_end(self, metadata_input: UTCDateTime, data_input: UTCDateTime) -> bool:
        if isinstance(metadata_input, UTCDateTime) and metadata_input >= data_input:
            return True
        else:
            logger.error("End date of {} should be in UTC and greater than "
                         "the data acquisition end time."
                         .format(self.__class__.__name__.split('_')[0]))
            return False

    def validate_time_period_start(self, metadata_input: UTCDateTime, data_input: UTCDateTime) -> bool:
        if isinstance(metadata_input, UTCDateTime) and metadata_input <= data_input:
            return True
        else:
            logger.error("Start date of {} should be in UTC and lower than "
                         "the data acquisition start time."
                         .format(self.__class__.__name__.split('_')[0]))
            return False


@dataclass
class Survey(BaseSurvey):
    """This class handles all the metadata attributes at the survey level."""
    acquired_by_author: Optional[str] = field(default=None,
                                              metadata={'xml_id': ['comments', 'mt.survey.acquired_by.author'],
                                                        'required': False,
                                                        'gui': True})
    acquired_by_comments: Optional[str] = field(default=None,
                                                metadata={'xml_id': ['comments', 'mt.survey.acquired_by.comments'],
                                                          'required': False,
                                                          'gui': True})
    archive_network: Optional[str] = field(default=None,
                                           metadata={'required': True,
                                                     'gui': True})
    citation_dataset_doi: Optional[str] = field(default=None,
                                                metadata={'xml_id': ['identifiers'],
                                                          'required': True,
                                                          'gui': True})
    comments: Optional[str] = field(default=None,
                                    metadata={'xml_id': ['comments', 'mt.survey.comments'],
                                              'required': False,
                                              'gui': True})
    country: Optional[str] = field(default=None,
                                   metadata={'xml_id': ['comments', 'mt.survey.country'],
                                             'required': False,
                                             'gui': True})
    geographic_name: Optional[str] = field(default=None,
                                           metadata={'xml_id': ['comments', 'mt.survey.geographic_name'],
                                                     'required': True,
                                                     'gui': True})
    name: Optional[str] = field(default=None,
                                metadata={'xml_id': ['description'],
                                          'required': True,
                                          'gui': True})
    project: Optional[str] = field(default=None,
                                   metadata={'xml_id': ['comments', 'mt.survey.project'],
                                             'required': True,
                                             'gui': True})
    project_lead_author: Optional[str] = field(default=None,
                                               metadata={'xml_id': ['operators.contacts.names'],
                                                         'required': False,
                                                         'gui': True})
    project_lead_email: Optional[str] = field(default=None,
                                              metadata={'xml_id': ['operators.contacts.emails'],
                                                        'required': False,
                                                        'gui': True})
    project_lead_organization: Optional[str] = field(default=None,
                                                     metadata={'xml_id': ['operators.agency'],
                                                               'required': False,
                                                               'gui': True})
    summary: Optional[str] = field(default=None,
                                   metadata={'xml_id': ['comments', 'mt.survey.summary'],
                                             'required': False,
                                             'gui': True})

    @staticmethod
    def validate_archive_network(metadata_input: str) -> bool:
        try:
            valid = re.match(r"^\w{2}$", metadata_input)
        except TypeError:
            logger.error("The network code should be a string.")
            return False
        else:
            if valid is None:
                logger.error("The network code should be two alphanumeric "
                             "character long.")
            return bool(valid)

    @staticmethod
    def validate_citation_dataset_doi(metadata_input: str) -> bool:
        # Proper formatting for DOI?
        try:
            dois = str2list(metadata_input)
        except AttributeError:
            logger.error("The DOI number(s) should be a string.")
            return False
        else:
            if not all([is_valid_uri(doi) for doi in dois]):
                logger.error("Invalid DOI(s). The DOI number(s) provided by "
                             "the archive should be strings formatted as "
                             "follows: 'scheme: path'.")
                return False
            return True

    @staticmethod
    def validate_project_lead_email(metadata_input: str) -> bool:
        if metadata_input is not None:
            try:
                emails = str2list(metadata_input)
            except AttributeError:
                logger.error("The project lead email(s) should be a string.")
                return False
            return all([check_email_formatting(email) for email in emails])
        return True


@dataclass
class BaseStation(BaseMetadata):
    """
    This class handles the metadata attributes shared at the station and
    channel levels.
    """
    data_type: str = field(default='LPMT',
                           metadata={'xml_id': ['comments', 'mt.station.data_type'],
                                     'required': True,
                                     'gui': False})
    elevation: Optional[float] = field(default=None,
                                       metadata={'required': True,
                                                 'gui': True})
    latitude: Optional[float] = field(default=None,
                                      metadata={'required': True,
                                                'gui': True})
    longitude: Optional[float] = field(default=None,
                                       metadata={'required': True,
                                                 'gui': True})

    @staticmethod
    def validate_geographics(param: tuple, metadata_input: float) -> bool:
        geo, min_range, max_range, min_val, max_val, units, msg = param
        try:
            metadata_input = float(metadata_input)
        except (TypeError, ValueError):
            logger.error("{} should be a float. {}".format(geo, msg))
            return False
        else:
            if metadata_input < min_range or metadata_input > max_range:
                logger.error("Unexpected {0}! The {0} should be between {1}"
                             "{3} and {2}{3}.".format(geo, min_range,
                                                      max_range, units))
                return False
            elif abs(metadata_input) < min_val or abs(metadata_input) > max_val:
                logger.error("Unexpected {0}! Provided {0} should "
                             "roughly match the {0} recorded by the on-site "
                             "GPS.".format(geo))
                return False
            return True

    def validate_elevation(self, metadata_input: float, data_input: float) -> bool:
        min_elev, max_elev = [abs(data_input) * x for x in [0.99, 1.01]]  # reasonable?
        param = ("Station elevation", -500, 8500, min_elev, max_elev, 'm', '')
        return BaseStation.validate_geographics(param, metadata_input)

    def validate_latitude(self, metadata_input: float, data_input: float) -> bool:
        min_lat, max_lat = [abs(data_input) * x for x in [0.995, 1.005]]  # reasonable?
        param = ("Station latitude", -90, 90, min_lat, max_lat, '°', '')
        return BaseStation.validate_geographics(param, metadata_input)

    def validate_longitude(self, metadata_input: float, data_input: float) -> bool:
        min_lon, max_lon = [abs(data_input) * x for x in [0.995, 1.005]]  # reasonable?
        param = ("Station longitude", -180, 180, min_lon, max_lon, '°', '')
        return BaseStation.validate_geographics(param, metadata_input)


@dataclass
class Station_(BaseSurvey, BaseStation):
    """This class handles all the metadata attributes at the station level."""
    archive_id: Optional[str] = field(default=None,
                                      metadata={'required': True,
                                                'gui': True})
    comments: Optional[str] = field(default=None,
                                    metadata={'xml_id': ['comments', 'mt.station.comments'],
                                              'required': False,
                                              'gui': True})
    declination_model: Optional[str] = field(default=None,
                                             metadata={'xml_id': ['comments', 'mt.station.location.declination.model'],
                                                       'required': True,
                                                       'gui': True})
    declination_value: Optional[float] = field(default=None,
                                               metadata={'xml_id': ['comments', 'mt.station.location.declination.value'],  # noqa: E501
                                                         'required': True,
                                                         'gui': True})
    geographic_name: Optional[str] = field(default=None,
                                           metadata={'xml_id': ['site'],
                                                     'required': False,
                                                     'gui': True})
    orientation_method: Optional[str] = field(default=None,
                                              metadata={'xml_id': ['comments', 'mt.station.orientation.method'],
                                                        'required': False,
                                                        'gui': True})
    orientation_reference_frame: Optional[str] = field(default=None,
                                                       metadata={'xml_id': ['comments', 'mt.station.orientation.reference_frame'],  # noqa: E501
                                                                 'required': True,
                                                                 'gui': True})
    run_list: str = field(default='a',
                          metadata={'xml_id': ['comments', 'mt.station.run_list'],
                                    'required': True,
                                    'gui': False})
    software_author: str = field(default=config['software']['author'],
                                 metadata={'xml_id': ['comments', 'mt.station.provenance.software.author'],
                                           'required': False,
                                           'gui': False})
    software_name: str = field(default=config['software']['name'],
                               metadata={'xml_id': ['comments', 'mt.station.provenance.software.name'],
                                         'required': False,
                                         'gui': False})
    software_version: str = field(default=config['software']['version'],
                                  metadata={'xml_id': ['comments', 'mt.station.provenance.software.version'],
                                            'required': False,
                                            'gui': False})
    submitter_author: Optional[str] = field(default=None,
                                            metadata={'xml_id': ['comments', 'mt.station.provenance.submitter.author'],
                                                      'required': False,
                                                      'gui': True})
    submitter_email: Optional[str] = field(default=None,
                                           metadata={'xml_id': ['comments', 'mt.station.provenance.submitter.email'],
                                                     'required': False,
                                                     'gui': True})
    submitter_organization: Optional[str] = field(default=None,
                                                  metadata={'xml_id': ['comments', 'mt.station.provenance.submitter.organization'],  # noqa: E501
                                                            'required': False,
                                                            'gui': True})

    @staticmethod
    def validate_archive_id(metadata_input: str) -> bool:
        try:
            valid = re.match(r"^\w{3,5}$", metadata_input)
        except TypeError:
            logger.error("The station name should be a string.")
            return False
        else:
            if valid is None:
                logger.error("The station name should be between three and "
                             "five alphanumeric character long.")
            return bool(valid)

    def validate_declination_value(self, metadata_input: float) -> bool:
        param = ("Declination angle", -90, 90, -99999, 99999, '°', None)
        return BaseStation.validate_geographics(param, metadata_input)

    @staticmethod
    def validate_submitter_email(metadata_input: str) -> bool:
        if metadata_input is not None:
            return check_email_formatting(metadata_input)
        return True


@dataclass
class Run(BaseMetadata):
    """This class handles all the metadata attributes at the run level."""
    acquired_by_author: Optional[str] = field(default=None,
                                              metadata={'xml_id': ['comments', 'mt.run:a.acquired_by.author'],
                                                        'required': False,
                                                        'gui': True})
    acquired_by_comments: Optional[str] = field(default=None,
                                                metadata={'xml_id': ['comments', 'mt.run:a.acquired_by.comments'],
                                                          'required': False,
                                                          'gui': True})
    components_recorded: Optional[str] = field(default=None,
                                               metadata={'required': True,
                                                         'gui': True})
    comments: Optional[str] = field(default=None,
                                    metadata={'xml_id': ['comments', 'mt.run:a.comments'],
                                              'required': False,
                                              'gui': True})
    data_type: str = field(default='LPMT',
                           metadata={'xml_id': ['type'],
                                     'required': True,
                                     'gui': False})
    datalogger_type: str = field(default='type: long-period 32-bit',
                                 metadata={'xml_id': ['description'],
                                           'required': False,
                                           'gui': False})
    datalogger_manufacturer: str = field(default='LEMI LLC.',
                                         metadata={'xml_id': ['manufacturer'],
                                                   'required': True,
                                                   'gui': False})
    datalogger_model: str = field(default='LEMI-424',
                                  metadata={'xml_id': ['model'],
                                            'required': True,
                                            'gui': True})
    datalogger_serial_number: Optional[str] = field(default=None,
                                                    metadata={'xml_id': ['serial_number'],
                                                              'required': True,
                                                              'gui': True})
    metadata_by_author: Optional[str] = field(default=None,
                                              metadata={'xml_id': ['comments', 'mt.run:a.metadata_by.author'],
                                                        'required': False,
                                                        'gui': True})
    metadata_by_comments: Optional[str] = field(default=None,
                                                metadata={'xml_id': ['comments', 'mt.run:a.metadata_by.comments'],
                                                          'required': False,
                                                          'gui': True})
    resource_id: Optional[str] = field(default=None,
                                       metadata={'xml_id': ['resource_id'],
                                                 'required': True,
                                                 'gui': False})
    time_period_end: Optional[UTCDateTime] = field(default=None,
                                                   metadata={'xml_id': ['removal_date'],
                                                             'required': True,
                                                             'gui': True})
    time_period_start: Optional[UTCDateTime] = field(default=None,
                                                     metadata={'xml_id': ['installation_date'],
                                                               'required': True,
                                                               'gui': True})

    @property
    def run_id(self) -> str:
        return self.resource_id.split(':')[1]  # type: ignore

    def validate_components_recorded(self, metadata_input: str) -> bool:
        """
        The following channels: Te, Tf, Ui, Sn, Fq, Ce (Sn: Satellite number,
        Fq: GPS Fix quality, Ce: Clock error) are recorded by default.
        """
        try:
            components = str2list(metadata_input)
        except AttributeError:
            logger.error("The list of components recorded for run '{}' "
                         "should be a string.".format(self.run_id))
            return False
        else:
            if not all([x in VALID_COMPONENTS for x in components]):
                logger.error("Some of the listed components for run '{}' are "
                             "not valid. List of valid components: {}."
                             .format(self.run_id, VALID_COMPONENTS))
                return False
            return True

    def validate_datalogger_serial_number(self, metadata_input: str) -> bool:
        equipment = "data logger (run {})".format(self.run_id)
        return check_serial_number(metadata_input, equipment)

    def validate_time_period_end(self, metadata_input: UTCDateTime, data_input: UTCDateTime) -> bool:
        if not (isinstance(metadata_input, UTCDateTime) and metadata_input <= data_input):
            logger.error("End date of run {} should be in UTC and lower than "
                         "the data acquisition end time."
                         .format(self.run_id))
            return False
        return True

    def validate_time_period_start(self, metadata_input: UTCDateTime, data_input: UTCDateTime) -> bool:
        if not (isinstance(metadata_input, UTCDateTime) and metadata_input >= data_input):
            logger.error("Start date of run {} should be in UTC and greater "
                         "than the data acquisition start time."
                         .format(self.run_id))
            return False
        return True


@dataclass
class BaseChannel(BaseStation):
    """
    This class handles the metadata attributes shared at the electric, magnetic
    and auxiliary levels.
    """
    channel_name: Optional[str] = field(default=None,
                                        metadata={'required': True,
                                                  'gui': False})
    component: Optional[str] = field(default=None,
                                     metadata={'xml_id': ['alternate_code'],
                                               'required': True,
                                               'gui': False})
    data_type: str = field(default='GEOPHYSICAL',
                           metadata={'xml_id': ['types'],
                                     'required': True,
                                     'gui': False})
    depth: int = field(default=0,
                       metadata={'required': True,
                                 'gui': False})
    instrument_manufacturer: Optional[str] = field(default=None,
                                                   metadata={'required': True,
                                                             'gui': True})
    instrument_model: Optional[str] = field(default=None,
                                            metadata={'required': True,
                                                      'gui': True})
    instrument_specs: Optional[str] = field(default=None,
                                            metadata={'required': True,
                                                      'gui': True})
    instrument_type: Optional[str] = field(default=None,
                                           metadata={'required': False,
                                                     'gui': True})
    location_code: str = field(default='',
                               metadata={'required': True,
                                         'gui': False})
    measurement_azimuth: Optional[float] = field(default=None,
                                                 metadata={'xml_id': ['azimuth'],
                                                           'required': True,
                                                           'gui': True})
    measurement_tilt: Optional[float] = field(default=None,
                                              metadata={'xml_id': ['dip'],
                                                        'required': True,
                                                        'gui': True})
    run_id: Optional[str] = field(default=None,
                                  metadata={'required': True,
                                            'gui': False})
    sample_rate: float = field(default=SAMPLING_RATE,
                               metadata={'xml_id': ['sample_rate'],
                                         'required': True,
                                         'gui': True})

    def validate_measurement_azimuth(self, metadata_input: float) -> bool:
        geo, msg = ["Azimuth angle", '']
        if isinstance(self, Electric):
            geo = ("Azimuth angle of electrode pair (run '{}' and channel "
                   "number '{}')".format(self.run_id, self.channel_number))  # type: ignore
            msg = MSG_E_CHA
        if isinstance(self, Magnetic):
            geo = "Azimuth angle of fluxgate (run '{}')".format(self.run_id)
        # -99999 and 99999 are the min and max azimuth - placeholder for the validate_geographics method
        param = (geo, 0, 360, -99999, 99999, '°', msg)
        return BaseStation.validate_geographics(param, metadata_input)

    def validate_measurement_tilt(self, metadata_input: float) -> bool:
        geo, msg = ["Tilt angle", '']
        if isinstance(self, Electric):
            geo = ("Tilt angle of electrode pair (run '{}' and channel '{}')"
                   .format(self.run_id, self.channel_number))  # type: ignore
            msg = MSG_E_CHA
        if isinstance(self, Magnetic):
            geo = "Tilt angle of fluxgate (run '{}')".format(self.run_id)
        # -99999 and 99999 are the min and max tilt - placeholder for the validate_geographics method
        param = (geo, 0, 180, -99999, 99999, '°', msg)
        return BaseStation.validate_geographics(param, metadata_input)


@dataclass
class Electric(BaseChannel):
    """This class handles all the metadata attributes at the electric level."""
    channel_number: Optional[str] = field(default=None,
                                          metadata={'xml_id': ['comments', 'mt.electric.channel_number'],
                                                    'required': True,
                                                    'gui': True})
    contact_resistance_end: Optional[float] = field(default=None,
                                                    metadata={'xml_id': ['comments', 'mt.electric.contact_resistance.end'],  # noqa: E501
                                                              'required': False,
                                                              'gui': True})
    contact_resistance_start: Optional[float] = field(default=None,
                                                      metadata={'xml_id': ['comments', 'mt.electric.contact_resistance.start'],  # noqa: E501
                                                                'required': False,
                                                                'gui': True})
    dc_end: Optional[float] = field(default=None,
                                    metadata={'xml_id': ['comments', 'mt.electric.dc.end'],
                                              'required': False,
                                              'gui': True})
    dc_start: Optional[float] = field(default=None,
                                      metadata={'xml_id': ['comments', 'mt.electric.dc.start'],
                                                'required': False,
                                                'gui': True})
    dipole_length: Optional[float] = field(default=None,
                                           metadata={'required': True,
                                                     'gui': True})
    negative_electrode_direction: Optional[str] = field(default=None,
                                                        metadata={'required': True,
                                                                  'gui': True})
    negative_electrode_serial_number: Optional[str] = field(default=None,
                                                            metadata={'required': True,
                                                                      'gui': True})
    positive_electrode_direction: Optional[str] = field(default=None,
                                                        metadata={'required': True,
                                                                  'gui': True})
    positive_electrode_serial_number: Optional[str] = field(default=None,
                                                            metadata={'required': True,
                                                                      'gui': True})

    def set_electric_component_channel(self) -> None:
        """
        Set component and channel name for a given electric channel based on
        the direction of its electrode pair.
        """
        try:
            direction = "-".join([self.positive_electrode_direction, self.negative_electrode_direction])  # type: ignore
        except TypeError:
            logger.error("The direction of the positive and negative electrodes "
                         "should be either: North, South, East or West. Please, "
                         "provide electrode direction(s) for run '{}' and "
                         "channel number '{}'! {}"
                         .format(self.run_id, self.channel_number, MSG_E_CHA))
            self.metadata_invalid.update({'positive_electrode_direction',
                                          'negative_electrode_direction'})
        else:
            component = E_COMPONENT_NAMING_CONVENTION.get(direction)
            if component is None:
                if self.positive_electrode_direction == self.negative_electrode_direction:
                    logger.error("The direction of the positive and negative "
                                 "electrodes in a given pair cannot be the "
                                 "same (check run '{}' and channel number "
                                 "'{}')!".format(self.run_id, self.channel_number))
                logger.warning("By convention, one electrode pair is "
                               "installed in a north-south direction and the "
                               "other pair in a  east-west direction (check "
                               "run '{}' and channel number '{}')!"
                               .format(self.run_id, self.channel_number))
                self.metadata_invalid.update({'positive_electrode_direction',
                                              'negative_electrode_direction'})
            else:
                self.component = component
                self.channel_name = CHANNEL_NAMING_CONVENTION[component]
                self.metadata_invalid.discard('positive_electrode_direction')
                self.metadata_invalid.discard('negative_electrode_direction')

    def set_electrode_info(self) -> None:
        """
        Set electrode manufacturer, model and type based on user provided
        electrode specs.
        """
        if self.instrument_specs == ELECTRODE_DEFAULT:
            self.instrument_manufacturer = "Borin"
            self.instrument_model = "STELTH 4"
            self.instrument_type = "Silver-Silver Chloride"
        elif self.instrument_specs is not None:
            specs = [x.split(':')[-1].strip() for x in self.instrument_specs.split('-')]
            self.instrument_manufacturer, self.instrument_model, self.instrument_type = specs

    def validate_instrument_specs(self, metadata_input: str) -> bool:
        equipment = 'electrode'
        channel_info = (" for run '{}' and channel number '{}'"
                        .format(self.run_id, self.channel_number))
        return check_instrument_specs(metadata_input, equipment, channel_info)

    def validate_e_property(self, param: tuple, metadata_input: float) -> bool:
        e_prop, min_range, max_range, units = param
        try:
            metadata_input = float(metadata_input)
        except (TypeError, ValueError):
            logger.error("The {} for run '{}' and channel number '{}' should "
                         "be a float. {}".format(e_prop, self.run_id,
                                                 self.channel_number, MSG_E_CHA))
            return False
        else:
            if metadata_input < min_range:
                logger.error("The {} for run '{}' and for channel number "
                             "'{}' should be positive."
                             .format(e_prop, self.run_id, self.channel_number))
                return False
            if metadata_input > max_range:
                msg = ''
                if 'resistance' in e_prop or 'voltage' in e_prop:
                    msg = ("A {} > {}{} is indicative of poor contact."
                           .format(e_prop, max_range, units))
                logger.warning("The {} for run '{}' and for channel number "
                               "'{}' should be less than {}{}. {}"
                               .format(e_prop, self.run_id, self.channel_number,
                                       max_range, units, msg))
            return True

    def validate_contact_resistance_end(self, metadata_input: float) -> bool:
        if metadata_input is not None:
            param = ("contact resistance (end)", 0, 3000, 'Ω')
            return self.validate_e_property(param, metadata_input)
        return True

    def validate_contact_resistance_start(self, metadata_input: float) -> bool:
        if metadata_input is not None:
            param = ("contact resistance (start)", 0, 3000, 'Ω')
            return self.validate_e_property(param, metadata_input)
        return True

    def validate_dc_end(self, metadata_input: float) -> bool:
        if metadata_input is not None:
            param = ("DC voltage (end)", 0, 0.1, 'V')
            return self.validate_e_property(param, metadata_input)
        return True

    def validate_dc_start(self, metadata_input: float) -> bool:
        if metadata_input is not None:
            param = ("DC voltage (start)", 0, 0.1, 'V')
            return self.validate_e_property(param, metadata_input)
        return True

    def validate_dipole_length(self, metadata_input: float) -> bool:
        param = ("dipole length", 0, 99999, 'm')
        return self.validate_e_property(param, metadata_input)

    def validate_negative_electrode_serial_number(self, metadata_input: str) -> bool:
        equipment = ("negative electrode (run '{}' and channel number '{}')"
                     .format(self.run_id, self.channel_number))
        return check_serial_number(metadata_input, equipment)

    def validate_positive_electrode_serial_number(self, metadata_input: str) -> bool:
        equipment = ("positive electrode (run '{}' and channel number: '{}')"
                     .format(self.run_id, self.channel_number))
        return check_serial_number(metadata_input, equipment)


@dataclass
class Magnetic(BaseChannel):
    """This class handles all the metadata attributes at the magnetic level."""
    serial_number: Optional[str] = field(default=None,
                                         metadata={'required': True,
                                                   'gui': True})

    def set_magnetic_info(self) -> None:
        if self.instrument_specs == FLUXGATE_DEFAULT:
            self.instrument_manufacturer = "LEMI LLC."
            self.instrument_model = "LEMI-039"
            self.instrument_type = "3-component analog magnetometer"
        elif self.instrument_specs is not None:
            specs = [x.split(':')[-1].strip() for x in self.instrument_specs.split('-')]
            self.instrument_manufacturer, self.instrument_model, self.instrument_type = specs

    def validate_instrument_specs(self, metadata_input: str) -> bool:
        equipment = "fluxgate"
        channel_info = " for run '{}'".format(self.run_id)
        return check_instrument_specs(metadata_input, equipment, channel_info)

    def validate_serial_number(self, metadata_input: str) -> bool:
        equipment = "fluxgate (run '{}')".format(self.run_id)
        return check_serial_number(metadata_input, equipment)


@dataclass
class Auxiliary(BaseChannel):
    """This class handles all the metadata attributes at the auxiliary level."""
    serial_number: Optional[str] = field(default=None,
                                         metadata={'required': True,
                                                   'gui': False})

    def set_auxiliary_info(self) -> None:
        self.instrument_specs = ""
        self.instrument_manufacturer = "LEMI LLC."
        self.instrument_model = "LEMI-424"
        self.instrument_type = "long-period 32-bit"

    def set_data_type(self):
        self.data_type = "HEALTH"

    def set_measurement_azimuth(self):
        self.measurement_azimuth = 0.0

    def set_measurement_tilt(self):
        self.measurement_tilt = 0.0


# -- Type aliases --
MD_VAL = Union[str, int, float, datetime, UTCDateTime]  # Type of metadata value
DCS = Union[Survey, Station_, Run, Electric, Magnetic, Auxiliary]  # Type of dataclasses
DCS_SUB_REMA = Union[Run, Electric, Magnetic, Auxiliary]
DCS_SUB_EMA = Union[Electric, Magnetic, Auxiliary]
DCS_SUB_MA = Union[Magnetic, Auxiliary]
WORKBOOK = openpyxl.Workbook
WORKSHEET = openpyxl.worksheet.worksheet.Worksheet
# -- Type aliases --


class LemiMetadata():
    """
    This class handles the parsing the LEMI field sheets, the population of the
    metadata and their validation.
    """

    def __init__(self, path2metadata: Path, output_xml: Path, lemi_data: LemiData) -> None:
        """Instantiate LemiMetadata class and scan metadata directory."""
        self.path2metadata = path2metadata
        self.output_xml = output_xml
        self.filenames: List[Path] = []
        self.categories: List[str] = ['Survey', 'Station', 'Run', 'Electric',
                                      'Magnetic', 'Auxiliary']
        self.survey: Survey = Survey()
        self.station: Station_ = Station_()
        self.run: List[Run] = []
        self.electric: List[Electric] = []
        self.magnetic: List[Magnetic] = []
        self.auxiliary: List[Auxiliary] = []
        self.data_stats: Dict = lemi_data.stats
        self.data_np: numpy.ndarray = lemi_data.data_np
        self.run_list: List[str] = get_run_list(lemi_data.stats['nbr_runs'])
        if path2metadata is not None:
            self.scan_path2metadata()

    def scan_path2metadata(self) -> None:
        """
        Scan metadata directory. If no field sheets are found under the given
        directory, the user will have to provide the necessary metadata in the
        GUI.
        """
        filenames = [tmp for tmp in self.path2metadata.glob('*.xlsx')
                     if not tmp.parts[-1].startswith('~$')]
        if not filenames:
            logger.error("No field sheet found under the following path - {}. "
                         "All metadata will have to be provided using the "
                         "GUI!".format(self.path2metadata))
        else:
            self.filenames = filenames

    @staticmethod
    def get_metadata_field(sheet: WORKSHEET, cell: str) -> Optional[MD_VAL]:
        """Get metadata field at a given cell in a given field sheet."""
        try:
            field_value = sheet[cell].value
        except Exception as error:
            logger.exception("Failed to retrieve value of cell {} - Skipping "
                             "cell - Exception: {}".format(cell, error))
            return None
        return field_value

    def identify_field_sheet(self, sheet: WORKSHEET, filename: Path) -> Optional[str]:
        """
        Check whether a field sheet is an install, removal or credential sheet.
        Layout between provided spread sheet templates varies and cell numbers
        listed in config file are organized in three sub-categories:
        install_sheet, credential_sheet, removal_sheet.
        """
        sheet_header = LemiMetadata.get_metadata_field(sheet, config['sheets']['header'])
        sheet_types = {val: key for key, val in config['sheets'].items()[1:]}
        sheet_type = sheet_types.get(sheet_header)
        if sheet_type is None:
            logger.warning("The following file {} does not have the proper "
                           "header. The provided spread sheet templates were "
                           "not used or their layout was modified. Skipping "
                           "file!".format(filename))
        return sheet_type

    @staticmethod
    def dataclass2dict(dc: DCS_SUB_REMA) -> Dict:
        """
        Convert dataclass instance to a dict and filter dict to:
        - only keep metadata fields displayed in GUI
        - reset missing and invalid metadata fields
        """
        missing_invalid = dc.metadata_missing | dc.metadata_invalid
        mdict = {}
        for mfield in fields(dc):
            if mfield.metadata.get('gui'):
                if mfield.name not in missing_invalid:
                    mdict[mfield.name] = getattr(dc, mfield.name)
                else:
                    mdict[mfield.name] = mfield.default
        return mdict

    def filter_channel(self, channel_type: str, run_id: str) -> List[DCS_SUB_EMA]:
        """
        For a given channel type (electric, magnetic or auxiliary), get list of
        channels with a given run id.
        """
        return [c for c in getattr(self, channel_type) if c.run_id == run_id]

    @staticmethod
    def update_channelnumber(channel: Electric, channel_number: str) -> Electric:
        """
        Make copy of populated Electric data classes and set channel number.
        """
        tmp = copy.deepcopy(channel)
        tmp.channel_number = channel_number
        return tmp

    @property
    def for_gui(self) -> Dict:
        """
        Define metadata output dictionary structure that will be used to set
        the model in the GUI.
        """
        output: Dict = {}
        for category in self.categories[:-1]:
            output[category] = {}
            mproperty = getattr(self, category.lower())
            if isinstance(mproperty, list):
                if category == 'Run':
                    if not mproperty:
                        mproperty = self.init_run_metadata_properties()
                    for ind, dc in enumerate(mproperty):
                        key = f'Run_{self.run_list[ind]}'
                        output[category][key] = LemiMetadata.dataclass2dict(dc)
                elif category == 'Electric':
                    for run_id in self.run_list:
                        channels = self.filter_channel('electric', run_id)
                        if not channels:
                            channel = self.init_channel_metadata_properties('electric', run_id)
                            channels = [LemiMetadata.update_channelnumber(channel, f'E{ind + 1}')  # type: ignore
                                        for ind in range(NBR_E_CHANNEL_MAX)]
                        for ind in range(NBR_E_CHANNEL_MAX):
                            key = f'Run_{run_id}_Electrode_Pair_{ind + 1}'
                            output[category][key] = LemiMetadata.dataclass2dict(channels[ind])
                elif category == 'Magnetic':
                    for run_id in self.run_list:
                        channels = self.filter_channel('magnetic', run_id)
                        if not channels:
                            dc = self.init_channel_metadata_properties('magnetic', run_id)
                        else:
                            dc = channels[0]
                        key = f'Run_{run_id}_Magnetic'
                        output[category][key] = LemiMetadata.dataclass2dict(dc)
            else:
                output[category] = LemiMetadata.dataclass2dict(mproperty)
        return output

    def parse_field_sheet(self, sheet: WORKSHEET, sheet_type: str, metadata_fields: Dict) -> Dict:
        """
        Parse field sheet using the cell numbers listed in the config file and
        return a nested dictionary of metadata entries organized based on the
        category under which a given metadata field falls.
        """
        logger.info("Parsing the {} {}".format(*sheet_type.split('_')))
        for category in self.categories[:-1]:
            if category not in metadata_fields:
                metadata_fields[category] = {}
            if sheet_type not in config[category]:
                continue
            logger.info("Parsing '{}' metadata".format(category))
            for key, val in config[category][sheet_type].items():
                if isinstance(val, list):
                    field_value = [LemiMetadata.get_metadata_field(sheet, x) for x in val]
                else:
                    field_value = LemiMetadata.get_metadata_field(sheet, val)  # type: ignore
                if key == 'comments':
                    field_value = ', '.join(filter(None, (metadata_fields[category].get(key, None),
                                                          field_value)))  # type: ignore
                metadata_fields[category][key] = field_value
        return metadata_fields

    def reformat_run(self, run_fields: Dict) -> Dict:
        """Reformat run metadata fields."""
        tmp = {'Run_{}'.format(run_id): dict.fromkeys(run_fields.keys())
               for run_id in self.run_list}
        tmp['Run_a'].update(run_fields)
        return tmp

    def reformat_electric(self, electric_fields: Dict) -> Dict:
        """Reformat electric metadata fields."""
        e_ids = get_e_ids(self.data_stats['nbr_runs'])
        tmp = {f'Run_{run_id}_Electrode_Pair_{ind}': dict.fromkeys(electric_fields.keys())
               for run_id, ind in e_ids}
        for key, val in electric_fields.items():
            for ind in range(1, NBR_E_CHANNEL_MAX + 1):
                if isinstance(val, list):
                    e_val = val[ind - 1]
                else:
                    e_val = val
                tmp[f'Run_a_Electrode_Pair_{ind}'][key] = e_val
        return tmp

    def reformat_magnetic(self, magnetic_fields: Dict) -> Dict:
        """Reformat magnetic metadata fields."""
        tmp = {f'Run_{run_id}_Magnetic': dict.fromkeys(magnetic_fields.keys())
               for run_id in self.run_list}
        tmp['Run_a_Magnetic'].update(magnetic_fields)
        return tmp

    def reformat_metadata_dict(self, metadata_fields: Dict) -> Dict:
        """
        Reformat metadata_fields to include all runs and associate correct
        metadata info to each electrode pair.
        """
        metadata_fields['Run'] = self.reformat_run(metadata_fields['Run'])
        metadata_fields['Electric'] = self.reformat_electric(metadata_fields['Electric'])
        metadata_fields['Magnetic'] = self.reformat_magnetic(metadata_fields['Magnetic'])
        return metadata_fields

    @property
    def from_field_sheets(self) -> Optional[Dict]:
        """Parse all field sheets provided by the user."""
        metadata_fields: Dict = {}
        parsed_sheets = []
        for filename in self.filenames:
            try:
                workbook = openpyxl.load_workbook(filename, data_only=True)  # type: ignore
            except Exception as error:
                logger.error("Failed to open field sheet: {} - Skipping file "
                             "- Exception: {}".format(filename, error))
            else:
                if workbook is None:
                    continue
                sheet = workbook.active
                sheet_type = self.identify_field_sheet(sheet, filename)
                if sheet_type is None:
                    continue
                if sheet_type not in parsed_sheets:
                    parsed_sheets.append(sheet_type)
                else:
                    logger.warning("Already parsed the {0} {1} - You may have more than one {0} "
                                   "{1} - Skipping {2}!".format(*sheet_type.split('_'), filename))
                    continue
                metadata_fields = self.parse_field_sheet(sheet, sheet_type, metadata_fields)
                workbook.close()
        return self.reformat_metadata_dict(metadata_fields) if not is_empty(metadata_fields) else None

    def init_run_metadata_properties(self) -> List[Run]:
        """
        Instantiate Run data class object and initialize non-user defined
        properties.
        """
        runs = []
        run_nbr = [x['run_nbr'] for x in self.data_np]
        starttime = [x['starttime'] for x in self.data_np]
        endtime = [x['endtime'] for x in self.data_np]
        for ind_run, run_id in enumerate(self.run_list, 1):
            run = Run(resource_id=':'.join(['mt.run.id', run_id]),
                      time_period_start=starttime[run_nbr.index(ind_run)],
                      time_period_end=endtime[len(run_nbr) - run_nbr[::-1].index(ind_run) - 1])
            runs.append(run)
        return runs

    def init_channel_metadata_properties(self, channel_type: str, run_id: str) -> DCS_SUB_EMA:
        """
        Create instances of Electric, Magnetic or Auxiliary data class and
        initialize properties inherited from the BaseSurvey and BaseStation
        data classes.
        """
        ChannelClass = globals()[channel_type.capitalize()]
        channel = ChannelClass(elevation=self.data_stats['elevation'],
                               latitude=self.data_stats['latitude'],
                               longitude=self.data_stats['longitude'],
                               run_id=run_id)
        return channel
